/*
 * Perform SPI communication using GPIO operations.
 *
 * Copyright (C) 2018, 2020, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/byteorder.h>
#include <l4/devices/spi-gpio.h>
#include <stdlib.h> /* NOTE: required by calloc/free */
#include <time.h>



Spi_gpio::Spi_gpio(Hw::Gpio_chip *clock_device, int clock_pin,
                   Hw::Gpio_chip *data_device, int data_pin,
                   Hw::Gpio_chip *enable_device, int enable_pin,
                   Hw::Gpio_chip *control_device, int control_pin)
: _clock_device(clock_device),
  _clock_pin(clock_pin),
  _data_device(data_device),
  _data_pin(data_pin),
  _enable_device(enable_device),
  _enable_pin(enable_pin),
  _control_device(control_device),
  _control_pin(control_pin)
{
  _clock_device->setup(_clock_pin, Hw::Gpio_chip::Output, 1);
  _data_device->setup(_data_pin, Hw::Gpio_chip::Output, 0);
  _enable_device->setup(_enable_pin, Hw::Gpio_chip::Output, 1);
  _control_device->setup(_control_pin, Hw::Gpio_chip::Output, 0);
}

Spi_gpio::Spi_gpio(uint64_t frequency,
                   Hw::Gpio_chip *clock_device, int clock_pin,
                   Hw::Gpio_chip *data_device, int data_pin,
                   Hw::Gpio_chip *enable_device, int enable_pin,
                   Hw::Gpio_chip *control_device, int control_pin)
: Spi_gpio(clock_device, clock_pin, data_device, data_pin, enable_device,
           enable_pin, control_device, control_pin)
{
  _frequency = frequency;
}

/* Send a byte sequence. */

uint32_t Spi_gpio::send(uint32_t bytes, const uint8_t data[])
{
  return send_dc(bytes, data, NULL, 8, false);
}

/* Send a byte sequence with accompanying data/command indicators. Bytes are
   grouped into character units, with each unit holding a value with the given
   byte ordering. */

uint32_t Spi_gpio::send_dc(uint32_t bytes, const uint8_t data[],
                           const int dc[], uint8_t char_size, bool big_endian)
{
  struct timespec ts;
  uint32_t offset, char_unit;
  uint8_t char_unit_size = ((char_size ? char_size - 1 : 0) / 8) + 1;

  if (_frequency)
  {
    ts.tv_sec = 0;
    ts.tv_nsec = 1000000000 / _frequency;
  }

  /* Initialise pin levels. */

  _enable_device->set(_enable_pin, 1);
  _clock_device->set(_clock_pin, 1);
  _data_device->set(_data_pin, 0);

  if ((_control_device != NULL) && (_control_pin >= 0) && (dc != NULL))
    _control_device->set(_control_pin, dc[0] ? 1 : 0);

  /* Enter the transmission state. */

  _enable_device->set(_enable_pin, 0);

  /* Clock data using the clock and data outputs. */

  for (offset = 0, char_unit = 0; offset < bytes; offset += char_unit_size, char_unit++)
  {
    uint32_t mask, value = get_stored_value(&data[offset], char_unit_size, big_endian);

    /* Set the control level, if appropriate. */

    if ((_control_device != NULL) && (_control_pin >= 0) && (dc != NULL))
      _control_device->set(_control_pin, dc[char_unit] ? 1 : 0);

    mask = 1 << (char_size - 1);

    for (uint8_t bit = 0; bit < char_size; bit++)
    {
      /* NOTE: Data presented on falling clock level and sampled on rising clock
               level. This is SPI mode 3, or 0 given that the enable level is
               driven low immediately before the first bit is presented. */

      _clock_device->set(_clock_pin, 0);
      _data_device->set(_data_pin, value & mask ? 1 : 0);

      if (_frequency)
        nanosleep(&ts, NULL);

      _clock_device->set(_clock_pin, 1);

      if (_frequency)
        nanosleep(&ts, NULL);

      mask >>= 1;
    }
  }

  _enable_device->set(_enable_pin, 1);

  return bytes;
}

/* Send a number of units, each holding a value limited to the given character
   size, with any bit immediately above the character portion of the unit
   indicating the data/command level. */

uint32_t Spi_gpio::send_units(uint32_t bytes, const uint8_t data[],
                              uint8_t unit_size, uint8_t char_size,
                              bool big_endian)
{
  uint32_t char_units = bytes / unit_size;
  uint8_t char_unit_size = ((char_size ? char_size - 1 : 0) / 8) + 1;

  /* Obtain the data/command values for each unit. */
  /* Obtain the actual character bytes to be sent. */

  /* NOTE: Work around stack allocation limitations in L4Re. We would want to
           use a static allocation as follows:

  int dc[char_units];
  uint8_t char_data[char_unit_size * char_units];
  */

  int *dc = (int *) calloc(char_units, sizeof(int));
  uint8_t *char_data = (uint8_t *) calloc(char_units, char_unit_size);

  /* Traverse the byte sequence, extracting data/command bits for each unit. */

  for (uint32_t offset = 0, char_unit = 0, char_offset = 0; char_unit < char_units;
       offset += unit_size, char_unit++, char_offset += char_unit_size)
  {
    /* Obtain the unit value. */

    uint32_t value = get_stored_value(&data[offset], unit_size, big_endian);

    /* The unit size must be greater than the character size for data/command
       bits to be present. */

    if (unit_size * 8 > char_size)
      dc[char_unit] = value & (1 << char_size) ? 1 : 0;

    /* Obtain the actual character data from the input data, discarding the most
       significant bytes beyond the character and masking out any remaining
       superfluous bits, thus obtaining a whole number of bytes. */

    value = value & ((1 << char_size) - 1);

    set_stored_value(value, &char_data[char_offset], char_unit_size, big_endian);
  }

  /* Discard the data/command values if they will not be used. */

  uint32_t result = send_dc(char_units * char_unit_size, char_data,
                            (unit_size * 8 <= char_size) ? NULL : dc,
                            char_size, big_endian);

  /* NOTE: Working around allocation limitations with dc... */

  free(dc);
  free(char_data);
  return result;
}

/* Perform a transfer, ignoring any DMA-specific parameters. */

uint32_t Spi_gpio::transfer(l4_addr_t vaddr,
                            l4re_dma_space_dma_addr_t paddr,
                            uint32_t count, uint8_t unit_size,
                            uint8_t char_size,
                            l4_addr_t desc_vaddr,
                            l4re_dma_space_dma_addr_t desc_paddr)
{
  (void) paddr; (void) desc_vaddr; (void) desc_paddr;

  /* Assume little endian data is being transferred, being the native data
     representation. */

  return send_units(count, (const uint8_t *) vaddr, unit_size, char_size,
                    false);
}



/* C language interface. */

void *spi_gpio_get_channel(uint64_t frequency,
                           void *clock_chip, int clock_pin,
                           void *data_chip, int data_pin,
                           void *enable_chip, int enable_pin,
                           void *control_chip, int control_pin)
{
  return (void *) new Spi_gpio(frequency,
                               reinterpret_cast<Hw::Gpio_chip *>(clock_chip), clock_pin,
                               reinterpret_cast<Hw::Gpio_chip *>(data_chip), data_pin,
                               reinterpret_cast<Hw::Gpio_chip *>(enable_chip), enable_pin,
                               reinterpret_cast<Hw::Gpio_chip *>(control_chip), control_pin);
}

uint32_t spi_gpio_send(void *channel, uint32_t bytes, const uint8_t data[])
{
  return static_cast<Spi_gpio *>(channel)->send(bytes, data);
}

uint32_t spi_gpio_send_dc(void *channel, uint32_t bytes, const uint8_t data[],
                          const int dc[], uint8_t char_size, int big_endian)
{
  return static_cast<Spi_gpio *>(channel)->send_dc(bytes, data, dc, char_size,
    big_endian);
}

uint32_t spi_gpio_send_units(void *channel, uint32_t bytes, const uint8_t data[],
                             uint8_t unit_size, uint8_t char_size, int big_endian)
{
  return static_cast<Spi_gpio *>(channel)->send_units(bytes, data, unit_size,
    char_size, big_endian);
}

uint32_t spi_gpio_transfer(void *channel, l4_addr_t vaddr,
                           l4re_dma_space_dma_addr_t paddr,
                           uint32_t count, uint8_t unit_size,
                           uint8_t char_size)
{ 
  return static_cast<Spi_gpio *>(channel)->transfer(vaddr, paddr, count,
    unit_size, char_size);
}

uint32_t spi_gpio_transfer_descriptor(void *channel, l4_addr_t vaddr,
                                      l4re_dma_space_dma_addr_t paddr,
                                      uint32_t count, uint8_t unit_size,
                                      uint8_t char_size, l4_addr_t desc_vaddr,
                                      l4re_dma_space_dma_addr_t desc_paddr)
{
  return static_cast<Spi_gpio *>(channel)->transfer(vaddr, paddr, count,
    unit_size, char_size, desc_vaddr, desc_paddr);    
}
